from lionagi.core.tool.tool_manager import func_to_tool
from typing import Callable
import lionagi.libs.ln_convert as convert
from .chat import MonoChat
from lionagi.core.schema.base_node import Tool
from lionagi.core.messages.schema import Instruction


class MonoReAct(MonoChat):

    async def ReAct(
        self,
        instruction: Instruction | str | dict[str, dict | str],
        context=None,
        sender=None,
        system=None,
        tools=None,
        num_rounds: int = 1,
        reason_prompt=None,
        action_prompt=None,
        output_prompt=None,
        **kwargs,
    ):
        return await self._ReAct(
            instruction,
            context=context,
            sender=sender,
            system=system,
            tools=tools,
            num_rounds=num_rounds,
            reason_prompt=reason_prompt,
            action_prompt=action_prompt,
            output_prompt=output_prompt,
            **kwargs,
        )

    @staticmethod
    def _get_prompt(prompt=None, default=None, num_steps=None, instruction=None):
        if prompt is not None:
            return prompt

        try:
            try:
                return default.format(num_steps=num_steps)
            except Exception:
                return default.format(instruction=instruction)
        except Exception:
            return default

    def _create_followup_config(self, tools, auto=True, **kwargs):

        if tools is not None and (
            isinstance(tools, list) and isinstance(tools[0], Tool)
        ):
            self.branch.tool_manager.register_tools(tools)

        if not self.branch.tool_manager.has_tools:
            raise ValueError("No tools found, You need to register tools")

        config = self.branch.tool_manager.parse_tool(tools=True, **kwargs)
        config["tool_parsed"] = True
        if auto:
            config["tool_choice"] = "auto"
        return config

    async def _handle_auto(
        self,
        _out=None,
        out=None,
        instruction=None,
        output_prompt=None,
        sender=None,
        output=False,
        config=None,
    ):
        if not self.branch._is_invoked():
            if output:
                return await self._handle_auto_output(
                    instruction, output_prompt, sender, out, config
                )
            return self._handle_auto_followup(_out, out)
        return False

    def _handle_auto_followup(self, _out, out):
        return _out if out else None

    async def _handle_auto_output(
        self, instruction, output_prompt, sender, out, config
    ):
        prompt_ = self._get_prompt(
            output_prompt, _output_prompt, instruction=instruction
        )
        return await self.chat(prompt_, sender=sender, out=out, **config)

    async def _ReAct(
        self,
        instruction: Instruction | str | dict[str, dict | str],
        context=None,
        sender=None,
        system=None,
        tools=None,
        num_rounds: int = 1,
        auto=False,
        reason_prompt=None,
        action_prompt=None,
        output_prompt=None,
        out=True,
        **kwargs,
    ):

        config = self._create_followup_config(tools, **kwargs)

        i = 0
        while i < num_rounds:
            prompt_ = self._get_prompt(
                reason_prompt, _reason_prompt, (num_rounds - i) * 2
            )
            instruct = {"Notice": prompt_}

            # reason
            if i == 0:
                instruct["Task"] = instruction

                await self.chat(
                    instruct,
                    context=context,
                    system=system,
                    sender=sender,
                    **kwargs,
                )

            elif i > 0:
                await self.chat(instruct, sender=sender, **kwargs)

            # action
            prompt_ = self._get_prompt(
                action_prompt, _action_prompt, (num_rounds - i) * 2 - 1
            )
            _out = await self.chat(prompt_, sender=sender, **config)

            if auto:
                a = await self._handle_auto(_out=_out, out=out)
                if a:
                    return a

            i += 1

        if auto:
            a = await self._handle_auto(
                instruction,
                output_prompt=output_prompt,
                sender=sender,
                out=out,
                output=True,
                config=config,
            )
            if a:
                return a

    async def _react(
        self,
        instruction=None,
        context=None,
        output_fields=None,
        tools=None,
        reason_prompt=None,
        action_prompt=None,
        **kwargs,
    ):

        config = self._create_followup_config(tools=tools, auto=False, **kwargs)

        instruct = {
            "requirement": "think step by step, perform reasoning and prepare action plan according to available tools only",
            "task": convert.to_str(instruction),
        }

        extra_fields = output_fields or {}
        output_fields = {
            "reason": reason_prompt or "reasoning",
            "action": action_prompt
            or "the action(s) to take, in function call format. If no actions are needed return none",
        }
        output_fields = {**output_fields, **extra_fields}

        out_ = await self.chat(
            instruct, context=context, output_fields=output_fields, **config
        )
        print(out_)

        if convert.strip_lower(out_["action"]) not in ["none", "no", "na", "nan"]:
            res = await self._invoke_tools(content_={"action_request": out_["action"]})
            for idx, item in enumerate(res):
                out_[f"action_response_{idx}"] = item

        return out_

    # TODO: auto_ReAct


_reason_prompt = """
you have {num_steps} step left in current task. if available, integrate previous tool responses. perform reasoning and prepare action plan according to available tools only, apply divide and conquer technique.
"""

_action_prompt = """
you have {num_steps} step left in current task, if further actions are needed, invoke tools usage. If you are done, present the final result to user without further tool usage
"""
_output_prompt = (
    "notice: present final output to user, original user instruction: {instruction}"
)



from pydantic import Field
from lionagi.libs import func_call
from ..prompt.prompt_template import ScoredTemplate
from ..branch import Branch


class PredictTemplate(ScoredTemplate):
    template_name: str = "default_predict_template"
    sentence: str | list | dict = Field(
        default_factory=str, description="the given sentence(s) to predict"
    )
    num_sentences: int = Field(
        default_factory=int, description="the number of sentences to predict"
    )
    answer: str | list = Field(
        default_factory=str, description="the predicted sentence(s)"
    )
    signature: str = "sentence -> answer"

    def __init__(
        self,
        sentence=None,
        num_sentences=None,
        confidence_score=False,
        reason=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.sentence = sentence
        self.num_sentences = num_sentences
        self.task = f"predict the next {self.num_sentences} sentence(s)"

        if reason:
            self.output_fields.append("reason")

        if confidence_score:
            self.output_fields.append("confidence_score")


async def predict(
    sentence,
    num_sentences=1,
    confidence_score=False,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    **kwargs,
):

    branch = Branch(
        name=branch_name,
        system=system,
        messages=messages,
        service=service,
        sender=sender,
        llmconfig=llmconfig,
        tools=tools,
        datalogger=datalogger,
        persist_path=persist_path,
        tool_manager=tool_manager,
    )

    predict_template = PredictTemplate(
        sentence=sentence,
        num_sentences=num_sentences,
        confidence_score=confidence_score,
        reason=reason,
    )

    await func_call.rcall(
        branch.chat,
        prompt_template=predict_template,
        retries=retries,
        delay=delay,
        backoff_factor=backoff_factor,
        default=default_value,
        timeout=timeout,
        **kwargs,
    )

    return predict_template



from lionagi.core.messages.schema import Instruction
from lionagi.core.schema.base_node import Tool
from .chat import MonoChat


class MonoFollowup(MonoChat):

    async def followup(
        self,
        instruction: Instruction | str | dict[str, dict | str],
        context=None,
        sender=None,
        system=None,
        tools=None,
        max_followup: int = 1,
        auto=False,
        followup_prompt=None,
        output_prompt=None,
        out=True,
        **kwargs,
    ):
        return await self._followup(
            instruction=instruction,
            context=context,
            sender=sender,
            system=system,
            tools=tools,
            max_followup=max_followup,
            auto=auto,
            followup_prompt=followup_prompt,
            output_prompt=output_prompt,
            out=out,
            **kwargs,
        )

    @staticmethod
    def _get_prompt(prompt=None, default=None, num_followup=None, instruction=None):
        if prompt is not None:
            return prompt

        try:
            try:
                return default.format(num_followup=num_followup)
            except Exception:
                return default.format(instruction=instruction)
        except Exception:
            return default

    def _create_followup_config(self, tools, **kwargs):

        if tools is not None and (
            isinstance(tools, list) and isinstance(tools[0], Tool)
        ):
            self.branch.tool_manager.register_tools(tools)

        if not self.branch.tool_manager.has_tools:
            raise ValueError("No tools found, You need to register tools")

        config = self.branch.tool_manager.parse_tool(tools=True, **kwargs)
        config["tool_parsed"] = True
        config["tool_choice"] = "auto"
        return config

    async def _handle_auto(
        self,
        _out=None,
        out=None,
        instruction=None,
        output_prompt=None,
        sender=None,
        output=False,
        **kwargs,
    ):
        if self.branch._is_invoked():
            return False

        if output:
            return await self._handle_auto_output(
                instruction, output_prompt, sender, out, **kwargs
            )
        return self._handle_auto_followup(_out, out)

    def _handle_auto_followup(self, _out, out):
        return _out if out else None

    async def _handle_auto_output(
        self, instruction, output_prompt, sender, out, **kwargs
    ):
        prompt = self._get_prompt(
            prompt=output_prompt, default=_output_prompt, instruction=instruction
        )
        return await self.chat(
            prompt, sender=sender, tool_parsed=True, out=out, **kwargs
        )

    async def _followup(
        self,
        instruction: Instruction | str | dict[str, dict | str],
        context=None,
        sender=None,
        system=None,
        tools=None,
        max_followup: int = 1,
        auto=False,
        followup_prompt=None,
        output_prompt=None,
        out=True,
        **kwargs,
    ) -> None:
        _out = ""
        config = self._create_followup_config(tools, **kwargs)

        i = 0
        while i < max_followup:

            prompt_ = self._get_prompt(
                prompt=followup_prompt,
                default=_followup_prompt,
                num_followup=max_followup - i,
            )

            if i == 0:

                prompt_ = {"notice": prompt_, "task": instruction}
                _out = await self.chat(
                    prompt_, context=context, system=system, sender=sender, **config
                )

            elif i > 0:
                _out = await self.chat(prompt_, sender=sender, **config)

                if auto:
                    a = self._handle_auto(_out=_out, out=out)
                    if a is not False:
                        return a

            i += 1

        if auto:
            a = await self._handle_auto(
                instruction=instruction,
                output_prompt=output_prompt,
                sender=sender,
                out=out,
                output=True,
                **kwargs,
            )
            if a:
                return a

        return _out


_followup_prompt = """In the current task you are allowed a maximum of another {num_steps} followup chats. if further actions are needed, invoke tools usage. If you are done, present the final result 
to user without further tool usage
"""

_output_prompt = (
    "notice: present final output to user, original user instruction: {instruction}"
)



import time
from typing import List, Any, Dict, Callable
from collections import deque
from pydantic import Field

from lionagi.libs import SysUtil, func_call, AsyncUtil

from .base_node import BaseRelatableNode, BaseNode, Tool
from lionagi.core.mail.schema import BaseMail

from lionagi.core.schema.condition import Condition

from lionagi.core.schema.action_node import ActionNode, ActionSelection
from lionagi.core.schema.base_node import Tool


class Relationship(BaseRelatableNode):
    """
    Represents a relationship between two nodes in a graph.

    Attributes:
            source_node_id (str): The identifier of the source node.
            target_node_id (str): The identifier of the target node.
            condition (Dict[str, Any]): A dictionary representing conditions for the relationship.

    Examples:
            >>> relationship = Relationship(source_node_id="node1", target_node_id="node2")
            >>> relationship.add_condition({"key": "value"})
            >>> condition_value = relationship.get_condition("key")
            >>> relationship.remove_condition("key")
    """

    source_node_id: str
    target_node_id: str
    bundle: bool = False
    condition: Callable = None

    def add_condition(self, condition: Condition):
        if not isinstance(condition, Condition):
            raise ValueError(
                "Invalid condition type, please use Condition class to build a valid condition"
            )
        self.condition = condition

    def check_condition(self, source_obj):
        try:
            return bool(self.condition(source_obj))
        except:
            raise ValueError("Invalid relationship condition function")

    def _source_existed(self, obj: Dict[str, Any]) -> bool:
        """
        Checks if the source node exists in a given object.

        Args:
                obj (Dict[str, Any]): The object to check.

        Returns:
                bool: True if the source node exists, False otherwise.
        """
        return self.source_node_id in obj.keys()

    def _target_existed(self, obj: Dict[str, Any]) -> bool:
        """
        Checks if the target node exists in a given object.

        Args:
                obj (Dict[str, Any]): The object to check.

        Returns:
                bool: True if the target node exists, False otherwise.
        """
        return self.target_node_id in obj.keys()

    def _is_in(self, obj: Dict[str, Any]) -> bool:
        """
        Validates the existence of both source and target nodes in a given object.

        Args:
                obj (Dict[str, Any]): The object to check.

        Returns:
                bool: True if both nodes exist.

        Raises:
                ValueError: If either the source or target node does not exist.
        """
        if self._source_existed(obj) and self._target_existed(obj):
            return True

        elif self._source_existed(obj):
            raise ValueError(f"Target node {self.source_node_id} does not exist")
        else:
            raise ValueError(f"Source node {self.target_node_id} does not exist")

    def __str__(self) -> str:
        """
        Returns a simple string representation of the Relationship.

        Examples:
                >>> relationship = Relationship(source_node_id="node1", target_node_id="node2")
                >>> str(relationship)
                'Relationship (id_=None, from=node1, to=node2, label=None)'
        """
        return (
            f"Relationship (id_={self.id_}, from={self.source_node_id}, to={self.target_node_id}, "
            f"label={self.label})"
        )

    def __repr__(self) -> str:
        """
        Returns a detailed string representation of the Relationship.

        Examples:
                >>> relationship = Relationship(source_node_id="node1", target_node_id="node2")
                >>> repr(relationship)
                'Relationship(id_=None, from=node1, to=node2, content=None, metadata=None, label=None)'
        """
        return (
            f"Relationship(id_={self.id_}, from={self.source_node_id}, to={self.target_node_id}, "
            f"content={self.content}, metadata={self.metadata}, label={self.label})"
        )


class Graph(BaseRelatableNode):
    """
    Represents a graph structure, consisting of nodes and their relationship.

    Attributes:
            nodes (Dict[str, BaseNode]): A dictionary of nodes in the graph.
            relationships (Dict[str, Relationship]): A dictionary of relationship between nodes in the graph.
            node_relationships (Dict[str, Dict[str, Dict[str, str]]]): A dictionary tracking the relationship of each node.

    Examples:
            >>> graph = Graph()
            >>> node = BaseNode(id_='node1')
            >>> graph.add_node(node)
            >>> graph.node_exists(node)
            True
            >>> relationship = Relationship(id_='rel1', source_node_id='node1', target_node_id='node2')
            >>> graph.add_relationship(relationship)
            >>> graph.relationship_exists(relationship)
            True
    """

    nodes: dict = Field(default={})
    relationships: dict = Field(default={})
    node_relationships: dict = Field(default={})

    def add_node(self, node: BaseNode) -> None:
        """
        Adds a node to the graph.

        Args:
                node (BaseNode): The node to add to the graph.
        """

        self.nodes[node.id_] = node
        self.node_relationships[node.id_] = {"in": {}, "out": {}}

    def add_relationship(self, relationship: Relationship) -> None:
        """
        Adds a relationship between nodes in the graph.

        Args:
                relationship (Relationship): The relationship to add.

        Raises:
                KeyError: If either the source or target node of the relationship is not found in the graph.
        """
        if relationship.source_node_id not in self.node_relationships.keys():
            raise KeyError(f"node {relationship.source_node_id} is not found.")
        if relationship.target_node_id not in self.node_relationships.keys():
            raise KeyError(f"node {relationship.target_node_id} is not found.")

        self.relationships[relationship.id_] = relationship
        self.node_relationships[relationship.source_node_id]["out"][
            relationship.id_
        ] = relationship.target_node_id
        self.node_relationships[relationship.target_node_id]["in"][
            relationship.id_
        ] = relationship.source_node_id

    def get_node_relationships(
        self, node: BaseNode = None, out_edge=True
    ) -> List[Relationship]:
        """
        Retrieves relationship of a specific node or all relationship in the graph.

        Args:
                node (Optional[BaseNode]): The node whose relationship to retrieve. If None, retrieves all relationship.
                out_edge (bool): Whether to retrieve outgoing relationship. If False, retrieves incoming relationship.

        Returns:
                List[Relationship]: A list of relationship.

        Raises:
                KeyError: If the specified node is not found in the graph.
        """
        if node is None:
            return list(self.relationships.values())

        if node.id_ not in self.nodes.keys():
            raise KeyError(f"node {node.id_} is not found")

        if out_edge:
            relationship_ids = list(self.node_relationships[node.id_]["out"].keys())
            relationships = func_call.lcall(
                relationship_ids, lambda i: self.relationships[i]
            )
            return relationships
        else:
            relationship_ids = list(self.node_relationships[node.id_]["in"].keys())
            relationships = func_call.lcall(
                relationship_ids, lambda i: self.relationships[i]
            )
            return relationships

    def get_predecessors(self, node: BaseNode):
        node_ids = list(self.node_relationships[node.id_]["in"].values())
        nodes = func_call.lcall(node_ids, lambda i: self.nodes[i])
        return nodes

    def get_successors(self, node: BaseNode):
        node_ids = list(self.node_relationships[node.id_]["out"].values())
        nodes = func_call.lcall(node_ids, lambda i: self.nodes[i])
        return nodes

    def remove_node(self, node: BaseNode) -> BaseNode:
        """
        Removes a node and its associated relationship from the graph.

        Args:
                node (BaseNode): The node to remove.

        Returns:
                BaseNode: The removed node.

        Raises:
                KeyError: If the node is not found in the graph.
        """
        if node.id_ not in self.nodes.keys():
            raise KeyError(f"node {node.id_} is not found")

        out_relationship = self.node_relationships[node.id_]["out"]
        for relationship_id, node_id in out_relationship.items():
            self.node_relationships[node_id]["in"].pop(relationship_id)
            self.relationships.pop(relationship_id)

        in_relationship = self.node_relationships[node.id_]["in"]
        for relationship_id, node_id in in_relationship.items():
            self.node_relationships[node_id]["out"].pop(relationship_id)
            self.relationships.pop(relationship_id)

        self.node_relationships.pop(node.id_)
        return self.nodes.pop(node.id_)

    def remove_relationship(self, relationship: Relationship) -> Relationship:
        """
        Removes a relationship from the graph.

        Args:
                relationship (Relationship): The relationship to remove.

        Returns:
                Relationship: The removed relationship.

        Raises:
                KeyError: If the relationship is not found in the graph.
        """
        if relationship.id_ not in self.relationships.keys():
            raise KeyError(f"relationship {relationship.id_} is not found")

        self.node_relationships[relationship.source_node_id]["out"].pop(
            relationship.id_
        )
        self.node_relationships[relationship.target_node_id]["in"].pop(relationship.id_)

        return self.relationships.pop(relationship.id_)

    def node_exist(self, node: BaseNode) -> bool:
        """
        Checks if a node exists in the graph.

        Args:
                node (BaseNode): The node to check.

        Returns:
                bool: True if the node exists, False otherwise.
        """
        if node.id_ in self.nodes.keys():
            return True
        else:
            return False

    def relationship_exist(self, relationship: Relationship) -> bool:
        """
        Checks if a relationship exists in the graph.

        Args:
                relationship (Relationship): The relationship to check.

        Returns:
                bool: True if the relationship exists, False otherwise.
        """
        if relationship.id_ in self.relationships.keys():
            return True
        else:
            return False

    def is_empty(self) -> bool:
        """
        Determines if the graph is empty.

        Returns:
                bool: True if the graph has no nodes, False otherwise.
        """
        if self.nodes:
            return False
        else:
            return True

    def clear(self) -> None:
        """Clears the graph of all nodes and relationship."""
        self.nodes.clear()
        self.relationships.clear()
        self.node_relationships.clear()

    def to_networkx(self, **kwargs) -> Any:
        """
        Converts the graph to a NetworkX graph object.

        Args:
                **kwargs: Additional keyword arguments to pass to the NetworkX DiGraph constructor.

        Returns:
                Any: A NetworkX directed graph representing the graph.

        Examples:
                >>> graph = Graph()
                >>> nx_graph = graph.to_networkx()
        """

        SysUtil.check_import("networkx")

        from networkx import DiGraph

        g = DiGraph(**kwargs)
        for node_id, node in self.nodes.items():
            node_info = node.to_dict()
            node_info.pop("node_id")
            node_info.update({"class_name": node.__class__.__name__})
            g.add_node(node_id, **node_info)

        for _, relationship in self.relationships.items():
            relationship_info = relationship.to_dict()
            relationship_info.pop("node_id")
            relationship_info.update({"class_name": relationship.__class__.__name__})
            source_node_id = relationship_info.pop("source_node_id")
            target_node_id = relationship_info.pop("target_node_id")
            g.add_edge(source_node_id, target_node_id, **relationship_info)

        return g


class Structure(BaseRelatableNode):
    graph: Graph = Graph()
    pending_ins: dict = {}
    pending_outs: deque = deque()
    execute_stop: bool = False
    condition_check_result: bool | None = None

    def add_node(self, node: BaseNode):
        self.graph.add_node(node)

    def add_relationship(
        self,
        from_node: BaseNode,
        to_node: BaseNode,
        bundle=False,
        condition=None,
        **kwargs,
    ):
        if isinstance(from_node, Tool) or isinstance(from_node, ActionSelection):
            raise ValueError(
                f"type {type(from_node)} should not be the head of the relationship, "
                f"please switch position and attach it to the tail of the relationship"
            )
        if isinstance(to_node, Tool) or isinstance(to_node, ActionSelection):
            bundle = True
        relationship = Relationship(
            source_node_id=from_node.id_,
            target_node_id=to_node.id_,
            bundle=bundle,
            **kwargs,
        )
        if condition:
            relationship.add_condition(condition)
        self.graph.add_relationship(relationship)

    def get_relationships(self) -> list[Relationship]:
        return self.graph.get_node_relationships()

    def get_node_relationships(self, node: BaseNode, out_edge=True, labels=None):
        relationships = self.graph.get_node_relationships(node, out_edge)
        if labels:
            if not isinstance(labels, list):
                labels = [labels]
            result = []
            for r in relationships:
                if r.label in labels:
                    result.append(r)
            relationships = result
        return relationships

    def get_predecessors(self, node: BaseNode):
        return self.graph.get_predecessors(node)

    def get_successors(self, node: BaseNode):
        return self.graph.get_successors(node)

    def node_exist(self, node: BaseNode) -> bool:
        return self.graph.node_exist(node)

    def relationship_exist(self, relationship: Relationship) -> bool:
        return self.graph.relationship_exist(relationship)

    def remove_node(self, node: BaseNode) -> BaseNode:
        return self.graph.remove_node(node)

    def remove_relationship(self, relationship: Relationship) -> Relationship:
        return self.graph.remove_relationship(relationship)

    def is_empty(self) -> bool:
        return self.graph.is_empty()

    def get_heads(self):
        heads = []
        for key in self.graph.node_relationships:
            if not self.graph.node_relationships[key]["in"]:
                heads.append(self.graph.nodes[key])
        return heads

    @staticmethod
    def parse_to_action(instruction: BaseNode, bundled_nodes: deque):
        action_node = ActionNode(instruction)
        while bundled_nodes:
            node = bundled_nodes.popleft()
            if isinstance(node, ActionSelection):
                action_node.action = node.action
                action_node.action_kwargs = node.action_kwargs
            elif isinstance(node, Tool):
                action_node.tools.append(node)
            else:
                raise ValueError("Invalid bundles nodes")
        return action_node

    async def check_condition(self, relationship, executable_id):
        if relationship.condition.source_type == "structure":
            return self.check_condition_structure(relationship)
        elif relationship.condition.source_type == "executable":
            self.send(
                recipient_id=executable_id, category="condition", package=relationship
            )
            while self.condition_check_result is None:
                await AsyncUtil.sleep(0.1)
                self.process_relationship_condition(relationship.id_)
                continue
            check_result = self.condition_check_result
            self.condition_check_result = None
            return check_result
        else:
            raise ValueError(f"Invalid source_type.")

    def check_condition_structure(self, relationship):
        return relationship.condition(self)

    async def get_next_step(self, current_node: BaseNode, executable_id):
        next_nodes = []
        next_relationships = self.get_node_relationships(current_node)
        for relationship in next_relationships:
            if relationship.bundle:
                continue
            if relationship.condition:
                check = await self.check_condition(relationship, executable_id)
                if not check:
                    continue
            node = self.graph.nodes[relationship.target_node_id]
            further_relationships = self.get_node_relationships(node)
            bundled_nodes = deque()
            for f_relationship in further_relationships:
                if f_relationship.bundle:
                    bundled_nodes.append(
                        self.graph.nodes[f_relationship.target_node_id]
                    )
            if bundled_nodes:
                node = self.parse_to_action(node, bundled_nodes)
            next_nodes.append(node)
        return next_nodes

    def acyclic(self):
        check_deque = deque(self.graph.nodes.keys())
        check_dict = {
            key: 0 for key in self.graph.nodes.keys()
        }  # 0: not visited, 1: temp, 2: perm

        def visit(key):
            if check_dict[key] == 2:
                return True
            elif check_dict[key] == 1:
                return False

            check_dict[key] = 1

            out_relationships = self.graph.get_node_relationships(self.graph.nodes[key])
            for node in out_relationships:
                check = visit(node.target_node_id)
                if not check:
                    return False

            check_dict[key] = 2
            return True

        while check_deque:
            key = check_deque.pop()
            check = visit(key)
            if not check:
                return False
        return True

    def send(self, recipient_id: str, category: str, package: Any) -> None:
        mail = BaseMail(
            sender_id=self.id_,
            recipient_id=recipient_id,
            category=category,
            package=package,
        )
        self.pending_outs.append(mail)

    def process_relationship_condition(self, relationship_id):
        for key in list(self.pending_ins.keys()):
            skipped_requests = deque()
            while self.pending_ins[key]:
                mail = self.pending_ins[key].popleft()
                if (
                    mail.category == "condition"
                    and mail.package["relationship_id"] == relationship_id
                ):
                    self.condition_check_result = mail.package["check_result"]
                else:
                    skipped_requests.append(mail)
            self.pending_ins[key] = skipped_requests

    async def process(self) -> None:
        for key in list(self.pending_ins.keys()):
            while self.pending_ins[key]:
                mail = self.pending_ins[key].popleft()
                if mail.category == "start":
                    next_nodes = self.get_heads()
                elif mail.category == "end":
                    self.execute_stop = True
                    return
                elif mail.category == "node_id":
                    if mail.package not in self.graph.nodes:
                        raise ValueError(
                            f"Node {mail.package} does not exist in the structure {self.id_}"
                        )
                    next_nodes = await self.get_next_step(
                        self.graph.nodes[mail.package], mail.sender_id
                    )
                elif mail.category == "node" and isinstance(mail.package, BaseNode):
                    if not self.node_exist(mail.package):
                        raise ValueError(
                            f"Node {mail.package} does not exist in the structure {self.id_}"
                        )
                    next_nodes = await self.get_next_step(mail.package, mail.sender_id)
                else:
                    raise ValueError(f"Invalid mail type for structure")

                if not next_nodes:  # tail
                    self.send(
                        recipient_id=mail.sender_id, category="end", package="end"
                    )
                else:
                    if len(next_nodes) == 1:
                        self.send(
                            recipient_id=mail.sender_id,
                            category="node",
                            package=next_nodes[0],
                        )
                    else:
                        self.send(
                            recipient_id=mail.sender_id,
                            category="node_list",
                            package=next_nodes,
                        )

    async def execute(self, refresh_time=1):
        if not self.acyclic():
            raise ValueError("Structure is not acyclic")

        while not self.execute_stop:
            await self.process()
            await AsyncUtil.sleep(refresh_time)



from typing import Any

from lionagi.libs import ln_convert as convert
from lionagi.libs.ln_async import AsyncUtil

from lionagi.core.messages.schema import Instruction
from lionagi.core.branch.branch import Branch

from lionagi.core.flow.base.baseflow import BasePolyFlow


class PolyChat(BasePolyFlow):

    def __init__(self, session) -> None:
        super().__init__(session)

    async def parallel_chat(
        self,
        instruction: Instruction | str,
        num_instances=1,
        context=None,
        sender=None,
        branch_system=None,
        messages=None,
        tools=False,
        out=True,
        invoke: bool = True,
        output_fields=None,
        persist_path=None,
        branch_config=None,
        explode=False,
        **kwargs,
    ) -> Any:
        """
        parallel chat
        """

        if branch_config is None:
            branch_config = {}
        return await self._parallel_chat(
            instruction,
            num_instances=num_instances,
            context=context,
            sender=sender,
            branch_system=branch_system,
            messages=messages,
            tools=tools,
            out=out,
            invoke=invoke,
            output_fields=output_fields,
            persist_path=persist_path,
            branch_config=branch_config,
            explode=explode,
            **kwargs,
        )

    async def _parallel_chat(
        self,
        instruction: Instruction | str,
        num_instances=1,
        context=None,
        sender=None,
        messages=None,
        tools=False,
        out=True,
        invoke: bool = True,
        output_fields=None,
        persist_path=None,
        branch_config={},
        explode=False,
        include_mapping=True,
        default_key="response",
        **kwargs,
    ) -> Any:
        """
        parallel chat
        """

        branches = {}

        async def _inner(i, ins_, cxt_):

            branch_ = Branch(
                messages=messages,
                service=self.session.default_branch.service,
                llmconfig=self.session.default_branch.llmconfig,
                persist_path=persist_path,
                **branch_config,
            )

            branch_.branch_name = branch_.id_

            if tools:
                branch_.tool_manager = self.session.default_branch.tool_manager

            res_ = await branch_.chat(
                instruction=ins_ or instruction,
                context=cxt_ or context,
                sender=sender,
                tools=tools,
                invoke=invoke,
                out=out,
                output_fields=output_fields,
                **kwargs,
            )

            branches[branch_.id_] = branch_
            if include_mapping:
                return {
                    "instruction": ins_ or instruction,
                    "context": cxt_ or context,
                    "branch_id": branch_.id_,
                    default_key: res_,
                }

            else:
                return res_

        async def _inner_2(i, ins_=None, cxt_=None):
            """returns num_instances of branches performing for same task/context"""
            tasks = [_inner(i, ins_, cxt_) for _ in range(num_instances)]
            ress = await AsyncUtil.execute_tasks(*tasks)
            return convert.to_list(ress)

        async def _inner_3(i):
            """different instructions but same context"""
            tasks = [_inner_2(i, ins_=ins_) for ins_ in convert.to_list(instruction)]
            ress = await AsyncUtil.execute_tasks(*tasks)
            return convert.to_list(ress)

        async def _inner_3_b(i):
            """different context but same instruction"""
            tasks = [_inner_2(i, cxt_=cxt_) for cxt_ in convert.to_list(context)]
            ress = await AsyncUtil.execute_tasks(*tasks)
            return convert.to_list(ress)

        async def _inner_4(i):
            """different instructions and different context"""

            tasks = []
            if explode:
                tasks = [
                    _inner_2(i, ins_=ins_, cxt_=cxt_)
                    for ins_ in convert.to_list(instruction)
                    for cxt_ in convert.to_list(context)
                ]
            else:
                tasks = [
                    _inner_2(i, ins_=ins_, cxt_=cxt_)
                    for ins_, cxt_ in zip(
                        convert.to_list(instruction), convert.to_list(context)
                    )
                ]

            ress = await AsyncUtil.execute_tasks(*tasks)
            return convert.to_list(ress)

        if len(convert.to_list(instruction)) == 1:
            if len(convert.to_list(context)) == 1:
                out_ = await _inner_2(0)
                self.session.branches.update(branches)
                return out_

            elif len(convert.to_list(context)) > 1:
                out_ = await _inner_3_b(0)
                self.session.branches.update(branches)
                return out_

        elif len(convert.to_list(instruction)) > 1:
            if len(convert.to_list(context)) == 1:
                out_ = await _inner_3(0)
                self.session.branches.update(branches)
                return out_

            elif len(convert.to_list(context)) > 1:
                out_ = await _inner_4(0)
                self.session.branches.update(branches)
                return out_



from lionagi.libs import convert, StringMatch


def check_number_field(x, fix_=True, **kwargs):
    if not isinstance(x, (int, float)):
        if fix_:
            try:
                return _fix_number_field(x, **kwargs)
            except Exception as e:
                raise e

        raise ValueError(
            f"Default value for NUMERIC must be an int or float, got {type(x).__name__}"
        )
    return x


def check_bool_field(x, fix_=True):
    if not isinstance(x, bool):
        if fix_:
            try:
                return _fix_bool_field(x)
            except Exception as e:
                raise e

        raise ValueError(
            f"Default value for BOOLEAN must be a bool, got {type(x).__name__}"
        )
    return x


def check_str_field(x, *args, fix_=True, **kwargs):
    if not isinstance(x, str):
        if fix_:
            try:
                return _fix_str_field(x, *args, **kwargs)
            except Exception as e:
                raise e

        raise ValueError(
            f"Default value for STRING must be a str, got {type(x).__name__}"
        )
    return x


def check_enum_field(x, model_fields_, choices, fix_=True, **kwargs):
    if "choices" not in model_fields_:
        raise ValueError(
            "Field type ENUM requires the 'choices' attribute to be specified."
        )

    same_dtype, dtype_ = convert.is_same_dtype(choices, return_dtype=True)
    if not same_dtype:
        raise ValueError(
            f"Field type ENUM requires all choices to be of the same type, got {choices}"
        )

    if not isinstance(x, dtype_):
        raise ValueError(
            f"Default value for ENUM must be an instance of the {dtype_.__name__}, got {type(x).__name__}"
        )

    if x not in choices:
        if fix_:
            try:
                return _fix_enum_field(x, choices, **kwargs)
            except Exception as e:
                raise e
        raise ValueError(
            f"Default value for ENUM must be one of the {choices}, got {x}"
        )

    return x


def _fix_number_field(x, *args, **kwargs):
    try:
        x = convert.to_str(x)
        x = convert.to_num(x, *args, **kwargs)
        if isinstance(x, (int, float)):
            return x
        raise ValueError(f"Failed to convert {x} into a numeric value")
    except Exception as e:
        raise ValueError(f"Failed to convert {x} into a numeric value") from e


def _fix_bool_field(x):
    try:
        x = convert.to_str(x)
        if (
            convert.strip_lower(x) in ["true", "1", "correct", "yes"]
            or convert.to_num(x) == 1
        ):
            return True
        elif (
            convert.strip_lower(x) in ["false", "0", "incorrect", "no", "none", "n/a"]
            or convert.to_num(x) == 0
        ):
            return False
        raise ValueError(f"Failed to convert {x} into a boolean value")
    except Exception as e:
        raise ValueError(f"Failed to convert {x} into a boolean value") from e


def _fix_str_field(x):
    try:
        x = convert.to_str(x)
        if isinstance(x, str):
            return x
        raise ValueError(f"Failed to convert {x} into a string value")
    except Exception as e:
        raise ValueError(f"Failed to convert {x} into a string value") from e


def _fix_enum_field(x, choices, **kwargs):
    try:
        x = convert.to_str(x)
        return StringMatch.choose_most_similar(x, choices, **kwargs)
    except Exception as e:
        raise ValueError(f"Failed to convert {x} into one of the choices") from e


validation_funcs = {
    "number": check_number_field,
    "bool": check_bool_field,
    "str": check_str_field,
    "enum": check_enum_field,
}



from typing import Any

from lionagi.core.flow.base.baseflow import BaseMonoFlow
from lionagi.core.flow.monoflow.chat_mixin import MonoChatMixin


class MonoChat(BaseMonoFlow, MonoChatMixin):

    def __init__(self, branch) -> None:
        super().__init__(branch)

    async def chat(
        self,
        instruction=None,
        context=None,
        sender=None,
        system=None,
        tools=False,
        out: bool = True,
        invoke: bool = True,
        output_fields=None,
        prompt_template=None,
        **kwargs,
    ) -> Any:
        """
        a chat conversation with LLM, processing instructions and system messages, optionally invoking tools.

        Args:
                branch: The Branch instance to perform chat operations.
                instruction (Union[Instruction, str]): The instruction for the chat.
                context (Optional[Any]): Additional context for the chat.
                sender (Optional[str]): The sender of the chat message.
                system (Optional[Union[System, str, Dict[str, Any]]]): System message to be processed.
                tools (Union[bool, Tool, List[Tool], str, List[str]]): Specifies tools to be invoked.
                out (bool): If True, outputs the chat response.
                invoke (bool): If True, invokes tools as part of the chat.
                **kwargs: Arbitrary keyword arguments for chat completion.

        Examples:
                >>> await ChatFlow.chat(branch, "Ask about user preferences")
        """

        config = self._create_chat_config(
            instruction=instruction,
            context=context,
            sender=sender,
            system=system,
            prompt_template=prompt_template,
            tools=tools,
            output_fields=output_fields,
            **kwargs,
        )

        await self._call_chatcompletion(**config)

        return await self._output(
            invoke=invoke,
            out=out,
            output_fields=output_fields,
            prompt_template=prompt_template,
        )



import contextlib
from collections import deque
from typing import Any

from lionagi.libs import convert, AsyncUtil, ParseUtil

from ..schema import BaseRelatableNode, ActionNode
from ..mail import BaseMail
from ..messages import System, Instruction
from ..agent import BaseAgent

from .branch import Branch


class ExecutableBranch(BaseRelatableNode):

    def __init__(self, verbose=True, **kwargs):
        super().__init__()
        self.branch: Branch = Branch(**kwargs)
        self.pending_ins = {}  # needed
        self.pending_outs = deque()  # needed
        self.responses = []
        self.execute_stop = False  # needed
        self.context = None  # needed
        self.context_log = []
        self.verbose = verbose

    def send(self, recipient_id: str, category: str, package: Any) -> None:
        mail = BaseMail(
            sender_id=self.id_,
            recipient_id=recipient_id,
            category=category,
            package=package,
        )
        self.pending_outs.append(mail)

    async def forward(self):

        for key in list(self.pending_ins.keys()):
            while self.pending_ins[key]:
                mail = self.pending_ins[key].popleft()
                if mail.category == "start":  # needed
                    self._process_start(mail)
                elif mail.category == "node":
                    await self._process_node(mail)
                elif mail.category == "node_list":
                    self._process_node_list(mail)
                elif mail.category == "condition":
                    self._process_condition(mail)
                elif mail.category == "end":  # needed
                    self._process_end(mail)

    async def execute(self, refresh_time=1):  # needed
        while not self.execute_stop:
            await self.forward()
            await AsyncUtil.sleep(refresh_time)

    async def _process_node(self, mail: BaseMail):

        if isinstance(mail.package, System):
            self._system_process(mail.package, verbose=self.verbose)
            self.send(mail.sender_id, "node_id", mail.package.id_)

        elif isinstance(mail.package, Instruction):
            await self._instruction_process(mail.package, verbose=self.verbose)
            self.send(mail.sender_id, "node_id", mail.package.id_)

        elif isinstance(mail.package, ActionNode):
            await self._action_process(mail.package, verbose=self.verbose)
            self.send(mail.sender_id, "node_id", mail.package.instruction.id_)
        else:
            try:
                await self._agent_process(mail.package, verbose=self.verbose)
                self.send(mail.sender_id, "node_id", mail.package.id_)
            except:
                raise ValueError(f"Invalid mail to process. Mail:{mail}")

    def _process_node_list(self, mail: BaseMail):
        self.send(mail.sender_id, "end", "end")
        self.execute_stop = True
        raise ValueError("Multiple path selection is currently not supported")

    def _process_condition(self, mail: BaseMail):
        relationship = mail.package
        check_result = relationship.condition(self)
        back_mail = {"relationship_id": mail.package.id_, "check_result": check_result}
        self.send(mail.sender_id, "condition", back_mail)

    def _system_process(self, system: System, verbose=True, context_verbose=False):
        from lionagi.libs import SysUtil

        SysUtil.check_import("IPython")
        from IPython.display import Markdown, display

        if verbose:
            print(f"------------------Welcome: {system.sender}--------------------")
            display(Markdown(f"system: {convert.to_str(system.system_info)}"))
            if self.context and context_verbose:
                display(Markdown(f"context: {convert.to_str(self.context)}"))

        self.branch.add_message(system=system)

    async def _instruction_process(
        self, instruction: Instruction, verbose=True, **kwargs
    ):
        from lionagi.libs import SysUtil

        SysUtil.check_import("IPython")
        from IPython.display import Markdown, display

        if verbose:
            display(
                Markdown(
                    f"{instruction.sender}: {convert.to_str(instruction.instruct)}"
                )
            )

        if self.context:
            instruction.content.update({"context": self.context})
            self.context = None

        result = await self.branch.chat(instruction, **kwargs)
        with contextlib.suppress(Exception):
            result = ParseUtil.fuzzy_parse_json(result)
            if "response" in result.keys():
                result = result["response"]
        if verbose and len(self.branch.assistant_responses) != 0:
            display(
                Markdown(
                    f"{self.branch.last_assistant_response.sender}: {convert.to_str(result)}"
                )
            )
            print("-----------------------------------------------------")

        self.responses.append(result)

    async def _action_process(self, action: ActionNode, verbose=True):
        from lionagi.libs import SysUtil

        SysUtil.check_import("IPython")
        from IPython.display import Markdown, display

        try:
            func = getattr(self.branch, action.action)
        except:
            raise ValueError(f"{action.action} is not a valid action")

        if verbose:
            display(
                Markdown(
                    f"{action.instruction.sender}: {convert.to_str(action.instruction.instruct)}"
                )
            )

        if action.tools:
            self.branch.register_tools(action.tools)
        if self.context:
            result = await func(
                action.instruction.content["instruction"],
                context=self.context,
                tools=action.tools,
                **action.action_kwargs,
            )
            self.context = None
        else:
            result = await func(
                action.instruction.content, tools=action.tools, **action.action_kwargs
            )

        if verbose and len(self.branch.assistant_responses) != 0:
            display(
                Markdown(
                    f"{self.branch.last_assistant_response.sender}: {convert.to_str(result)}"
                )
            )
            print("-----------------------------------------------------")

        self.responses.append(result)

    async def _agent_process(self, agent, verbose=True):
        context = self.responses
        if verbose:
            print("*****************************************************")
        result = await agent.execute(context)

        if verbose:
            print("*****************************************************")

        self.context = result
        self.responses.append(result)

    def _process_start(self, mail):
        start_mail_content = mail.package
        self.context = start_mail_content["context"]
        self.send(start_mail_content["structure_id"], "start", "start")

    def _process_end(self, mail):
        self.execute_stop = True
        self.send(mail.sender_id, "end", "end")



from collections import deque
from typing import Any, Union, TypeVar, Callable

from lionagi.libs.sys_util import PATH_TYPE
from lionagi.libs import StatusTracker, BaseService, convert, dataframe

from ..schema import TOOL_TYPE, Tool, DataLogger
from ..tool import ToolManager, func_to_tool

from ..messages import System
from ..mail import BaseMail

from .util import MessageUtil
from .base_branch import BaseBranch
from .branch_flow_mixin import BranchFlowMixin

from dotenv import load_dotenv

load_dotenv()

T = TypeVar("T", bound=Tool)


class Branch(BaseBranch, BranchFlowMixin):

    def __init__(
        self,
        name: str | None = None,
        system: dict | list | System | None = None,
        messages: dataframe.ln_DataFrame | None = None,
        service: BaseService | None = None,
        sender: str | None = None,
        llmconfig: dict[str, str | int | dict] | None = None,
        tools: list[Callable | Tool] | None = None,
        datalogger: None | DataLogger = None,
        persist_path: PATH_TYPE | None = None,  # instruction_sets=None,
        tool_manager: ToolManager | None = None,
        **kwargs,
    ):

        # init base class
        super().__init__(
            messages=messages,
            datalogger=datalogger,
            persist_path=persist_path,
            name=name,
            **kwargs,
        )

        # add branch name
        self.sender = sender or "system"

        # add tool manager and register tools
        self.tool_manager = tool_manager or ToolManager()
        if tools:
            try:
                tools_ = []
                _tools = convert.to_list(tools)
                for i in _tools:
                    if isinstance(i, Tool):
                        tools_.append(i)
                    else:
                        tools_.append(func_to_tool(i))

                self.register_tools(tools_)
            except Exception as e:
                raise TypeError(f"Error in registering tools: {e}") from e

        # add service and llmconfig
        self.service, self.llmconfig = self._add_service(service, llmconfig)
        self.status_tracker = StatusTracker()

        # add instruction sets
        # self.instruction_sets = instruction_sets

        # add pending ins and outs for mails
        self.pending_ins = {}
        self.pending_outs = deque()

        # add system
        if system is not None:
            self.add_message(system=system)

    @classmethod
    def from_csv(
        cls,
        filepath,
        name: str | None = None,
        service: BaseService | None = None,
        llmconfig: dict[str, str | int | dict] | None = None,
        tools: TOOL_TYPE | None = None,
        datalogger: None | DataLogger = None,
        persist_path: PATH_TYPE | None = None,  # instruction_sets=None,
        tool_manager: ToolManager | None = None,
        read_kwargs=None,
        **kwargs,
    ):

        return cls._from_csv(
            filepath=filepath,
            read_kwargs=read_kwargs,
            name=name,
            service=service,
            llmconfig=llmconfig,
            tools=tools,
            datalogger=datalogger,
            persist_path=persist_path,
            # instruction_sets=instruction_sets,
            tool_manager=tool_manager,
            **kwargs,
        )

    @classmethod
    def from_json_string(
        cls,
        filepath,
        name: str | None = None,
        service: BaseService | None = None,
        llmconfig: dict[str, str | int | dict] | None = None,
        tools: TOOL_TYPE | None = None,
        datalogger: None | DataLogger = None,
        persist_path: PATH_TYPE | None = None,  # instruction_sets=None,
        tool_manager: ToolManager | None = None,
        read_kwargs=None,
        **kwargs,
    ):

        return cls._from_json(
            filepath=filepath,
            read_kwargs=read_kwargs,
            name=name,
            service=service,
            llmconfig=llmconfig,
            tools=tools,
            datalogger=datalogger,
            persist_path=persist_path,
            # instruction_sets=instruction_sets,
            tool_manager=tool_manager,
            **kwargs,
        )

    def messages_describe(self) -> dict[str, Any]:

        return dict(
            total_messages=len(self.messages),
            summary_by_role=self._info(),
            summary_by_sender=self._info(use_sender=True),
            # instruction_sets=self.instruction_sets,
            registered_tools=self.tool_manager.registry,
            messages=[msg.to_dict() for _, msg in self.messages.iterrows()],
        )

    @property
    def has_tools(self) -> bool:
        return self.tool_manager.registry != {}

    # todo: also update other attributes
    def merge_branch(self, branch: "Branch", update: bool = True) -> None:

        message_copy = branch.messages.copy()
        self.messages = self.messages.merge(message_copy, how="outer")
        self.datalogger.extend(branch.datalogger.log)

        if update:
            # self.instruction_sets.update(branch.instruction_sets)
            self.tool_manager.registry.update(branch.tool_manager.registry)
        else:
            for key, value in branch.instruction_sets.items():
                if key not in self.instruction_sets:
                    self.instruction_sets[key] = value

            for key, value in branch.tool_manager.registry.items():
                if key not in self.tool_manager.registry:
                    self.tool_manager.registry[key] = value

    # ----- tool manager methods ----- #
    def register_tools(self, tools: Union[Tool, list[Tool]]) -> None:

        if not isinstance(tools, list):
            tools = [tools]
        self.tool_manager.register_tools(tools=tools)

    def delete_tools(
        self,
        tools: Union[T, list[T], str, list[str]],
        verbose: bool = True,
    ) -> bool:

        if not isinstance(tools, list):
            tools = [tools]
        if convert.is_same_dtype(tools, str):
            for act_ in tools:
                if act_ in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_)
            if verbose:
                print("tools successfully deleted")
            return True
        elif convert.is_same_dtype(tools, Tool):
            for act_ in tools:
                if act_.schema_["function"]["name"] in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_.schema_["function"]["name"])
            if verbose:
                print("tools successfully deleted")
            return True
        if verbose:
            print("tools deletion failed")
        return False

    def send(self, recipient: str, category: str, package: Any) -> None:

        mail_ = BaseMail(
            sender=self.sender, recipient=recipient, category=category, package=package
        )
        self.pending_outs.append(mail_)

    def receive(
        self,
        sender: str,
        messages: bool = True,
        tools: bool = True,
        service: bool = True,
        llmconfig: bool = True,
    ) -> None:

        skipped_requests = deque()
        if sender not in self.pending_ins:
            raise ValueError(f"No package from {sender}")
        while self.pending_ins[sender]:
            mail_ = self.pending_ins[sender].popleft()

            if mail_.category == "messages" and messages:
                if not isinstance(mail_.package, dataframe.ln_DataFrame):
                    raise ValueError("Invalid messages format")
                MessageUtil.validate_messages(mail_.package)
                self.messages = self.messages.merge(mail_.package, how="outer")

            elif mail_.category == "tools" and tools:
                if not isinstance(mail_.package, Tool):
                    raise ValueError("Invalid tools format")
                self.tool_manager.register_tools([mail_.package])

            elif mail_.category == "provider" and service:
                from lionagi.libs.ln_api import BaseService

                if not isinstance(mail_.package, BaseService):
                    raise ValueError("Invalid provider format")
                self.service = mail_.package

            elif mail_.category == "llmconfig" and llmconfig:
                if not isinstance(mail_.package, dict):
                    raise ValueError("Invalid llmconfig format")
                self.llmconfig.update(mail_.package)

            else:
                skipped_requests.append(mail_)

        self.pending_ins[sender] = skipped_requests
        if self.pending_ins[sender] == deque():
            self.pending_ins.pop(sender)

    def receive_all(self) -> None:

        for key in list(self.pending_ins.keys()):
            self.receive(key)

    @staticmethod
    def _add_service(service, llmconfig):
        from lionagi.integrations.provider.oai import OpenAIService

        if service is None:
            try:
                from lionagi.integrations.provider import Services

                service = Services.OpenAI()

            except:
                raise ValueError("No available service")
        if llmconfig is None:
            if isinstance(service, OpenAIService):
                from lionagi.integrations.config import oai_schema

                llmconfig = oai_schema["chat/completions"]["config"]
            else:
                llmconfig = {}
        return service, llmconfig

    def _is_invoked(self) -> bool:
        """
        Check if the conversation has been invoked with an action response.

        Returns:
                bool: True if the conversation has been invoked, False otherwise.

        """
        content = self.messages.iloc[-1]["content"]
        try:
            if convert.to_dict(content)["action_response"].keys() >= {
                "function",
                "arguments",
                "output",
            }:
                return True
        except Exception:
            return False



from pydantic import Field

from lionagi.libs import func_call, convert
from ..prompt.prompt_template import ScoredTemplate
from ..branch import Branch


class ScoreTemplate(ScoredTemplate):
    template_name: str = "default_score"
    sentence: str | list | dict = Field(
        default_factory=str, description="the given context to score"
    )
    answer: float = Field(default_factory=float, description=f"a numeric score")
    signature: str = "sentence -> answer"

    def __init__(
        self,
        sentence=None,
        instruction=None,
        score_range=(1, 10),
        inclusive=True,
        num_digit=0,
        confidence_score=False,
        reason=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.sentence = sentence

        return_precision = ""
        if num_digit == 0:
            return_precision = "integer"
        else:
            return_precision = f"num:{convert.to_str(num_digit)}f"

        self.task = f"""
score context according to the following constraints
1. objective, {convert.to_str(instruction)}
2. score range, {convert.to_str(score_range)}
3. include_endpoints, {"yes" if inclusive else "no"}
4. format the score in {return_precision}
"""

        if reason:
            self.output_fields.append("reason")

        if confidence_score:
            self.output_fields.append("confidence_score")

        self.out_validation_kwargs['answer'] = {
            "upper_bound": score_range[1],
            "lower_bound": score_range[0],
            "num_type": int if num_digit == 0 else float,
            "precision": num_digit if num_digit != 0 else None,
        }

async def score(
    sentence,
    instruction=None,
    score_range=(1, 10),
    inclusive=True,
    num_digit=0,
    confidence_score=False,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    **kwargs,
):

    branch = Branch(
        name=branch_name,
        system=system,
        messages=messages,
        service=service,
        sender=sender,
        llmconfig=llmconfig,
        tools=tools,
        datalogger=datalogger,
        persist_path=persist_path,
        tool_manager=tool_manager,
    )

    _template = ScoreTemplate(
        sentence=sentence,
        instruction=instruction,
        score_range=score_range,
        inclusive=inclusive,
        num_digit=num_digit,
        confidence_score=confidence_score,
        reason=reason,
    )

    await func_call.rcall(
        branch.chat,
        prompt_template=_template,
        retries=retries,
        delay=delay,
        backoff_factor=backoff_factor,
        default=default_value,
        timeout=timeout,
        **kwargs,
    )

    return _template



from enum import Enum
from pydantic import Field

from lionagi.libs import func_call, StringMatch
from ..prompt.prompt_template import ScoredTemplate
from ..branch import Branch


class SelectTemplate(ScoredTemplate):
    template_name: str = "default_select"
    sentence: str | list | dict = Field(default_factory=str, description="the given context")
    answer: Enum | str = Field(default_factory=str, description="selection from given choices")
    
    signature: str = "sentence -> answer"

    def __init__(
        self,
        sentence=None,
        choices=None,
        num_choices=1,
        instruction=None,
        reason=False, 
        confidence_score=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.sentence = sentence
        self.choices = choices
        self.task = f"select {num_choices} item(s), from provided choices {choices}."
        if instruction:
            self.task += f"objetive {instruction}."
        
        if reason:
            self.output_fields.append("reason")

        if confidence_score:
            self.output_fields.append("confidence_score")


async def select(
    sentence,
    choices=None,
    num_choices=1,
    instruction=None,
    confidence_score=False,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    **kwargs,
):

    branch = Branch(
        name=branch_name,
        system=system,
        messages=messages,
        service=service,
        sender=sender,
        llmconfig=llmconfig,
        tools=tools,
        datalogger=datalogger,
        persist_path=persist_path,
        tool_manager=tool_manager,
    )

    _template = SelectTemplate(
        sentence=sentence,
        choices=choices,
        num_choices=num_choices,
        instruction=instruction,
        confidence_score=confidence_score,
        reason=reason,
    )

    await func_call.rcall(
        branch.chat,
        prompt_template=_template,
        retries=retries,
        delay=delay,
        backoff_factor=backoff_factor,
        default=default_value,
        timeout=timeout,
        **kwargs,
    )

    ans = _template.answer
    if ans not in _template.choices:
        _template.answer = StringMatch.choose_most_similar(ans, _template.choices)

    return _template



from lionagi.libs import convert, func_call, StringMatch
from lionagi.core.schema.base_node import BaseComponent
from datetime import time, datetime, date

from pydantic import BaseModel, Field
from .field_validator import validation_funcs


_non_prompt_words = [
    "id_",
    "node_id",
    "meta",
    "timestamp",
    "metadata",
    "signature",
    "task",
    "template_name",
]


class PromptTemplate(BaseComponent):
    signature: str = Field("null", description="signature indicating inputs, outputs")

    def __init__(
        self,
        template_name: str = "default_prompt_template",
        version_: str | float | int = None,
        description_: dict | str | None = None,
        task_: str | None = None,
        choices_: list[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.template_name = template_name
        self.meta_insert(["version"], version_)
        self.meta_insert(["description"], description_ or "")
        self.task = task_
        self.choices = choices_
        self.output_fields = None
        self.input_fields, self.output_fields = self._get_input_output_fields(
            self.signature
        )
        self.out_validation_kwargs = {}

    @property
    def version(self):
        return self.metadata["version"]

    @version.setter
    def version(self, value):
        self.metadata["version"] = value

    @property
    def description(self):
        return self.metadata["description"]

    @description.setter
    def description(self, value):
        self.metadata["description"] = value

    @property
    def prompt_fields(self):
        return [
            _field for _field in self.property_keys if _field not in _non_prompt_words
        ]

    @staticmethod
    def _get_input_output_fields(str_):
        _inputs, _outputs = str_.split("->")

        _inputs = [convert.strip_lower(i) for i in _inputs.split(",")]
        _outputs = [convert.strip_lower(o) for o in _outputs.split(",")]

        return _inputs, _outputs

    @property
    def instruction_context(self):
        a = "".join(
            f"""
        ## input: {i}:
        - description: {self.model_fields[i].description}
        - value: {str(self.__getattribute__(self.input_fields[idx]))}
        """
            for idx, i in enumerate(self.input_fields)
        )
        return a.replace("        ", "")

    @property
    def instruction(self):
        ccc = f"""
        0. Your task is {self.task},
        1. provided: {self.input_fields}, 
        2. requested: {self.output_fields}
        ----------
        """
        return ccc.replace("        ", "")

    @property
    def instruction_output_fields(self):
        return {i: self.model_fields[i].description for i in self.output_fields}

    @property
    def prompt_fields_annotation(self):
        dict_ = {i: self.model_fields[i].annotation for i in self.prompt_fields}
        for k, v in dict_.items():
            if "|" in str(v):
                v = str(v)
                v = v.split("|")
                dict_[k] = func_call.lcall(v, convert.strip_lower)
            else:
                dict_[k] = [v.__name__]

        return dict_

    def _validate_field(self, k, v, fix_=True, **kwargs):
        str_ = self.prompt_fields_annotation[k]

        if "enum" in str_:
            self.__setattr__(
                k,
                validation_funcs["enum"](v, choices=self.choices, fix_=fix_, **kwargs),
            )
            return True

        elif "bool" in str_:
            self.__setattr__(k, validation_funcs["bool"](v))
            return True

        elif "int" in str_ or "float" in str_:
            self.__setattr__(k, validation_funcs["number"](v, fix_=fix_, **kwargs))
            return True

        elif "str" in str_:
            self.__setattr__(k, validation_funcs["str"](v, fix_=fix_, **kwargs))
            return True

        return False


    def _process_response(self, out_, fix_=True):
        kwargs = self.out_validation_kwargs.copy()
        for k, v in out_.items():
            if k not in kwargs:
                kwargs = {k: {}}
            if self._validate_field(k, v, fix_=fix_, **kwargs[k]):
                continue
            else:
                raise ValueError(f"field {k} with value {v} is not valid")

    @property
    def out(self):
        return {i: self.__getattribute__(i) for i in self.output_fields}



class ScoredTemplate(PromptTemplate):
    confidence_score: float | None = Field(
        default_factory=float,
        description="a numeric score between 0 to 1 formatted in num:0.2f",
    )
    reason: str | None = Field(
        default_factory=str, description="brief reason for the given output"
    )

# class Weather(PromptTemplate):
#     sunny: bool = Field(True, description="true if the weather is sunny outside else false")
#     rainy: bool = Field(False, description="true if it is raining outside else false")
#     play1: bool = Field(True, description="conduct play1")
#     play2: bool = Field(False, description="conduct play2")
#     signature: str = Field("sunny, rainy -> play1, play2")

# predictor = Weather(
#     template_name="predictor",
#     task_ = "decides to conduct one of play1 and play2",
#     version_=0.1,
#     description_="predicts the weather and decides play",
#     signature = "sunny, play1 -> play2"
#     )

# predictor.to_instruction()
